/* slider.cpp */
/* Pavel Surynek */

#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <time.h>
#include <malloc.h>
#include <limits.h>
#include <math.h>

#include <map>
#include <set>


typedef struct
{
	int m_V;
	int *m_vrtxs;
	int *m_goal;

	int m_C1;
	int *m_cycle1;

	int m_C2;
	int *m_cycle2;

	int m_C3;
	int *m_cycle3;

	int m_H1;
	int *m_hash1;

	int m_H2;
	int *m_hash2;

	int m_H3;
	int *m_hash3;
} SLIDER;


typedef struct
{
	int m_V;
	int **m_dist;
} DIST_MATRIX;


typedef enum 
{
	MOVE_UNDEFINED,
	CYCLE_1_LEFT,
	CYCLE_1_RIGHT,
	CYCLE_2_LEFT,
	CYCLE_2_RIGHT,
	CYCLE_3_LEFT,
	CYCLE_3_RIGHT
} MOVE;


typedef struct
{
	int m_M, m_maxM;
	MOVE *m_moves;
} SOLUTION;


typedef struct
{
	int m_c1_sign;
	int m_c2_sign;
} BI_SIGNATURE;


typedef struct
{
	int m_c1_sign;
	int m_c2_sign;
	int m_c3_sign;
} TRI_SIGNATURE;


class QUAD_SIGNATURE
{
public:
	bool operator==(const QUAD_SIGNATURE &q_sig) const
	{
		return (   m_c1_sign == q_sig.m_c1_sign
			&& m_c2_sign == q_sig.m_c2_sign
			&& m_c3_sign == q_sig.m_c3_sign
			&& m_c4_sign == q_sig.m_c4_sign);
	}

	bool operator<(const QUAD_SIGNATURE &q_sig) const
	{
		return ((m_c1_sign < q_sig.m_c1_sign) ? true : ((m_c1_sign > q_sig.m_c1_sign) ? false :
			((m_c2_sign < q_sig.m_c2_sign) ? true : ((m_c2_sign > q_sig.m_c2_sign) ? false :
			 ((m_c3_sign < q_sig.m_c3_sign) ? true : ((m_c3_sign > q_sig.m_c3_sign) ? false :
			  ((m_c4_sign < q_sig.m_c4_sign) ? true : ((m_c4_sign > q_sig.m_c4_sign) ? false : false))))))));
	}

public:
	int m_c1_sign;
	int m_c2_sign;
	int m_c3_sign;
	int m_c4_sign;
};


class STAR_NODE
{
public:
	bool operator==(const STAR_NODE &star) const
	{
		return (m_estimated == star.m_estimated);
	}

	bool operator<(const STAR_NODE &star) const
	{
		return ((m_estimated < star.m_estimated) ? true : ((m_estimated < star.m_estimated) ? false : false));
	}

public:
	int m_estimated;
	SLIDER m_slider;
};


typedef std::set<STAR_NODE, std::less<STAR_NODE> > STAR_LIST_set;
STAR_LIST_set STAR_OPEN_LIST;


typedef std::map<QUAD_SIGNATURE, int, std::less<QUAD_SIGNATURE> > QUAD_SIGN_DEPTHS_map;
QUAD_SIGN_DEPTHS_map QUAD_SIGN_DEPTHS;
QUAD_SIGN_DEPTHS_map QUAD_SIGN_R_DEPTHS;
QUAD_SIGN_DEPTHS_map QUAD_SIGN_RR_DEPTHS;


SLIDER ORIGINAL_1;
SLIDER ORIGINAL_2;


#define MAX_DEPTHS_COUNT    2048

int MAX_DEPTHS[MAX_DEPTHS_COUNT];


#define SIGNS_COUNT     256

//int O_SIGNS[SIGNS_COUNT];
//int G_SIGNS[SIGNS_COUNT];
int SIGN_DEPTHS[SIGNS_COUNT];


#define BI_SIGNS_COUNT    256

//int O_BI_SIGNS[BI_SIGNS_COUNT][BI_SIGNS_COUNT];
//int G_BI_SIGNS[BI_SIGNS_COUNT][BI_SIGNS_COUNT];
int BI_SIGN_DEPTHS[BI_SIGNS_COUNT][BI_SIGNS_COUNT];


#define TRI_SIGNS_COUNT    256

//int O_TRI_SIGNS[TRI_SIGNS_COUNT][TRI_SIGNS_COUNT];
//int G_TRI_SIGNS[TRI_SIGNS_COUNT][TRI_SIGNS_COUNT];
int TRI_SIGN_DEPTHS[TRI_SIGNS_COUNT][TRI_SIGNS_COUNT][TRI_SIGNS_COUNT];


long STEPS;


int calc_signature(int *init, int *goal, int V)
{
	int trans = 0;
	int i, j, k, l;

	for (i = 0; i < V; ++i)
	{
		for (j = 0; j < V; ++j)
		{
			if (goal[j] == init[i])
			{
				break;
			}
		}

		for (k = 0; k < i; ++k)
		{
			for (l = j + 1; l < V; ++l)
			{
				if (init[k] == goal[l])
				{
					++trans;
				}
			}
		}
	}

	return trans;
}


void calc_bi_signature(SLIDER *init, SLIDER *goal, BI_SIGNATURE *bi_sign)
{
	int trans_1 = 0;
	int trans_2 = 0;
	int i, j, k, l;

	for (i = 0; i < init->m_V; ++i)
	{
		for (j = 0; j < init->m_V; ++j)
		{
			if (goal->m_goal[j] == init->m_vrtxs[i])
			{
				break;
			}
		}

		for (k = 0; k < i; ++k)
		{
			for (l = j + 1; l < init->m_V; ++l)
			{
				if (init->m_vrtxs[k] == goal->m_goal[l])
				{
					++trans_1;
				}
			}
		}
	}

	bi_sign->m_c1_sign = trans_1;

	for (i = 0; i < init->m_H1; ++i)
	{
		for (j = 0; j < init->m_H1; ++j)
		{
			if (goal->m_goal[init->m_hash1[j]] == init->m_vrtxs[init->m_hash1[i]])
			{
				break;
			}
		}

		for (k = 0; k < i; ++k)
		{
			for (l = j + 1; l < init->m_H1; ++l)
			{
				if (init->m_vrtxs[init->m_hash1[k]] == goal->m_goal[init->m_hash1[l]])
				{
					++trans_2;
				}
			}
		}
	}

	bi_sign->m_c2_sign = trans_2;
}


void calc_tri_signature(SLIDER *init, SLIDER *goal, TRI_SIGNATURE *tri_sign)
{
	int trans_1 = 0;
	int trans_2 = 0;
	int trans_3 = 0;
	int i, j, k, l;

	for (i = 0; i < init->m_V; ++i)
	{
		for (j = 0; j < init->m_V; ++j)
		{
			if (goal->m_goal[j] == init->m_vrtxs[i])
			{
				break;
			}
		}

		for (k = 0; k < i; ++k)
		{
			for (l = j + 1; l < init->m_V; ++l)
			{
				if (init->m_vrtxs[k] == goal->m_goal[l])
				{
					++trans_1;
				}
			}
		}
	}

	tri_sign->m_c1_sign = trans_1;

	for (i = 0; i < init->m_H1; ++i)
	{
		for (j = 0; j < init->m_H1; ++j)
		{
			if (goal->m_goal[init->m_hash1[j]] == init->m_vrtxs[init->m_hash1[i]])
			{
				break;
			}
		}

		for (k = 0; k < i; ++k)
		{
			for (l = j + 1; l < init->m_H1; ++l)
			{
				if (init->m_vrtxs[init->m_hash1[k]] == goal->m_goal[init->m_hash1[l]])
				{
					++trans_2;
				}
			}
		}
	}

	tri_sign->m_c2_sign = trans_2;

	for (i = 0; i < init->m_H2; ++i)
	{
		for (j = 0; j < init->m_H2; ++j)
		{
			if (goal->m_goal[init->m_hash2[j]] == init->m_vrtxs[init->m_hash2[i]])
			{
				break;
			}
		}

		for (k = 0; k < i; ++k)
		{
			for (l = j + 1; l < init->m_H2; ++l)
			{
				if (init->m_vrtxs[init->m_hash2[k]] == goal->m_goal[init->m_hash2[l]])
				{
					++trans_3;
				}
			}
		}
	}

	tri_sign->m_c3_sign = trans_3;
}


void calc_quad_signature(SLIDER *init, SLIDER *goal, QUAD_SIGNATURE *quad_sign)
{
	int trans_1 = 0;
	int trans_2 = 0;
	int trans_3 = 0;
	int trans_4 = 0;
	int i, j, k, l;

	for (i = 0; i < init->m_V; ++i)
	{
		for (j = 0; j < init->m_V; ++j)
		{
			if (goal->m_goal[j] == init->m_vrtxs[i])
			{
				break;
			}
		}

		for (k = 0; k < i; ++k)
		{
			for (l = j + 1; l < init->m_V; ++l)
			{
				if (init->m_vrtxs[k] == goal->m_goal[l])
				{
					++trans_1;
				}
			}
		}
	}

	quad_sign->m_c1_sign = trans_1;

	for (i = 0; i < init->m_H1; ++i)
	{
		for (j = 0; j < init->m_H1; ++j)
		{
			if (goal->m_goal[init->m_hash1[j]] == init->m_vrtxs[init->m_hash1[i]])
			{
				break;
			}
		}

		for (k = 0; k < i; ++k)
		{
			for (l = j + 1; l < init->m_H1; ++l)
			{
				if (init->m_vrtxs[init->m_hash1[k]] == goal->m_goal[init->m_hash1[l]])
				{
					++trans_2;
				}
			}
		}
	}

	quad_sign->m_c2_sign = trans_2;

	for (i = 0; i < init->m_H2; ++i)
	{
		for (j = 0; j < init->m_H2; ++j)
		{
			if (goal->m_goal[init->m_hash2[j]] == init->m_vrtxs[init->m_hash2[i]])
			{
				break;
			}
		}

		for (k = 0; k < i; ++k)
		{
			for (l = j + 1; l < init->m_H2; ++l)
			{
				if (init->m_vrtxs[init->m_hash2[k]] == goal->m_goal[init->m_hash2[l]])
				{
					++trans_3;
				}
			}
		}
	}

	quad_sign->m_c3_sign = trans_3;

	for (i = 0; i < init->m_H3; ++i)
	{
		for (j = 0; j < init->m_H3; ++j)
		{
			if (goal->m_goal[init->m_hash3[j]] == init->m_vrtxs[init->m_hash3[i]])
			{
				break;
			}
		}

		for (k = 0; k < i; ++k)
		{
			for (l = j + 1; l < init->m_H3; ++l)
			{
				if (init->m_vrtxs[init->m_hash3[k]] == goal->m_goal[init->m_hash3[l]])
				{
					++trans_4;
				}
			}
		}
	}

	quad_sign->m_c4_sign = trans_4;
}


void print_signatures(int V)
{
	int max = V * (V - 1) / 2;
	int i;

	printf("Signatures:\n");

	printf("O:");
	for (i = 0; i < max; ++i)
	{
//		printf("%d ", O_SIGNS[i]);
	}
	printf("\n");

	printf("G:");
	for (i = 0; i < max; ++i)
	{
//		printf("%d ", G_SIGNS[i]);
	}
	printf("\n");

	printf("D:");
	for (i = 0; i < max; ++i)
	{
		if (SIGN_DEPTHS[i] != INT_MAX)
		{
			printf("%d ", SIGN_DEPTHS[i]);
		}
		else
		{
			printf("- ");
		}
	}
	printf("\n");

}


void print_bi_signatures(int V)
{
	int max = V * (V - 1) / 2;
	int i, j;

	printf("Bi-Signatures:\n");

	printf("O:\n");
	for (i = 0; i < max; ++i)
	{
		for (j = 0; j < max; ++j)
		{
//			printf("%d ", O_BI_SIGNS[i][j]);
		}
		printf("\n");
	}
	printf("\n");

	printf("G:\n");
	for (i = 0; i < max; ++i)
	{
		for (j = 0; j < max; ++j)
		{
//			printf("%d ", G_BI_SIGNS[i][j]);
		}
		printf("\n");
	}
	printf("\n");

	printf("D:\n");
	for (i = 0; i < max; ++i)
	{
		for (j = 0; j < max; ++j)
		{
			if (BI_SIGN_DEPTHS[i][j] != INT_MAX)
			{
				printf("%d ", BI_SIGN_DEPTHS[i][j]);
			}
			else
			{
				printf("- ");
			}
		}
		printf("\n");
	}
	printf("\n");
}


double analyze_tri_signatures(int V)
{
	int max = V * (V - 1) / 2;
	int i, j, k;

	long init_cnt = 0;
	long total_cnt = 0;

	for (i = 0; i < max; ++i)
	{
		for (j = 0; j < max; ++j)
		{
			for (k = 0; k < max; ++k)
			{
				if (TRI_SIGN_DEPTHS[i][j][k] != INT_MAX)
				{
					++init_cnt;
					++total_cnt;
				}
				else
				{
					++total_cnt;
				}
			}
		}
	}

	printf(" Init:%ld\n", init_cnt);
	printf("Total:%ld\n", total_cnt);

	return ((double)init_cnt / (double)total_cnt);
}


void mult_dist_matrix(DIST_MATRIX *dm)
{
	int i, j, k;

	for (i = 0; i < dm->m_V; ++i)
	{
		for (j = 0; j < dm->m_V; ++j)
		{
			for (k = 0; k < dm->m_V; ++k)
			{
				int dist;

				if (dm->m_dist[i][k] == INT_MAX)
				{
					if (dm->m_dist[k][j] == INT_MAX)
					{
						dist = INT_MAX;
					}
					else
					{
						dist = INT_MAX;
					}
				}
				else
				{
					if (dm->m_dist[k][j] == INT_MAX)
					{
						dist = INT_MAX;
					}
					else
					{
						dist = dm->m_dist[i][k] + dm->m_dist[k][j];
					}
				}

				if (dm->m_dist[i][j] > dist)
				{
					dm->m_dist[i][j] = dist;
				}
			}
		}
	}
}


void print_dist_matrix(DIST_MATRIX *dm)
{
	int i, j;
	printf("Distance matrix:%d\n", dm->m_V);

	for (i = 0; i < dm->m_V; ++i)
	{
		for (j = 0; j < dm->m_V; ++j)
		{
			printf("%d\t", dm->m_dist[i][j]);
		}
		printf("\n");
	}
}


void init_dist_matrix(SLIDER *sl, DIST_MATRIX *dm)
{
	int i, j;

	dm->m_V = sl->m_V;

	dm->m_dist = (int**)malloc(sl->m_V * sizeof(int*));
	assert(dm->m_dist != NULL);

	for (i = 0; i < sl->m_V; ++i)
	{
		dm->m_dist[i] = (int*)malloc(sl->m_V * sizeof(int));
		assert(dm->m_dist[i] != NULL);
	}

	for (i = 0; i < dm->m_V; ++i)
	{
		for (j = 0; j < dm->m_V; ++j)
		{
			dm->m_dist[i][j] = INT_MAX;
		}
	}

	for (i = 0; i < dm->m_V; ++i)
	{
		dm->m_dist[i][i] = 0;
	}

	for (i = 0; i < sl->m_C1; ++i)
	{
		dm->m_dist[sl->m_cycle1[i]][sl->m_cycle1[(i + 1) % sl->m_C1]] = 1;
		dm->m_dist[sl->m_cycle1[(i + 1) % sl->m_C1]][sl->m_cycle1[i]] = 1;
	}

	for (i = 0; i < sl->m_C2; ++i)
	{
		dm->m_dist[sl->m_cycle2[i]][sl->m_cycle2[(i + 1) % sl->m_C2]] = 1;
		dm->m_dist[sl->m_cycle2[(i + 1) % sl->m_C2]][sl->m_cycle2[i]] = 1;
	}

	for (i = 0; i < sl->m_C3; ++i)
	{
		dm->m_dist[sl->m_cycle3[i]][sl->m_cycle3[(i + 1) % sl->m_C3]] = 1;
		dm->m_dist[sl->m_cycle3[(i + 1) % sl->m_C3]][sl->m_cycle3[i]] = 1;
	}

	int log2 = (int)(log((double)dm->m_V) / log(2.0)) + 1;

	for (i = 0; i < log2; ++i)
	{
		mult_dist_matrix(dm);
	}
}


void init_slider(int c1_a, int c1_b, int c2_a, int c2_b, SLIDER *slider)
{
	slider->m_V = c2_b + 1;
	slider->m_vrtxs = (int*)malloc(slider->m_V * sizeof(int));
	assert(slider->m_vrtxs != NULL);

	slider->m_goal = (int*)malloc(slider->m_V * sizeof(int));
	assert(slider->m_goal != NULL);

	slider->m_C1 = c1_b + 1;
	slider->m_cycle1 = (int*)malloc(slider->m_C1 * sizeof(int)); 
	assert(slider->m_cycle1 != NULL);

	slider->m_C2 = c1_b - c2_a + c2_b - c1_b + 2;
	slider->m_cycle2 = (int*)malloc(slider->m_C2 * sizeof(int)); 
	assert(slider->m_cycle2 != NULL);

	slider->m_C3 = c2_b - c1_b + c2_a + 1;
	slider->m_cycle3 = (int*)malloc(slider->m_C3 * sizeof(int)); 
	assert(slider->m_cycle3 != NULL);

	slider->m_H1 = slider->m_V;
	slider->m_hash1 = (int*)malloc(slider->m_H1 * sizeof(int));
	assert(slider->m_hash1 != NULL);

	slider->m_H2 = slider->m_V;
	slider->m_hash2 = (int*)malloc(slider->m_H2 * sizeof(int));
	assert(slider->m_hash2 != NULL);

	slider->m_H3 = slider->m_V;
	slider->m_hash3 = (int*)malloc(slider->m_H3 * sizeof(int));
	assert(slider->m_hash3 != NULL);

	int i, j;

	for (i = 0; i < slider->m_C1; ++i)
	{
		slider->m_cycle1[i] = i;
	}

	slider->m_cycle2[0] = 0;

	int n1 = c1_b - c2_a;

	for (i = 0; i <= n1; ++i)
	{
		slider->m_cycle2[i + 1] = c1_b - i;
	}

	for (j = 0; j <= c2_a; ++j)
	{
		slider->m_cycle3[j] = j;
	}

	while (c1_b < c2_b)
	{
		++i;
		++c1_b;
		slider->m_cycle2[i] = c1_b;
		slider->m_cycle3[j] = c1_b;
		++j;
	}

	for (i = 0; i < slider->m_V; ++i)
	{
		slider->m_vrtxs[i] = i;
		slider->m_goal[i] = i;
		slider->m_hash1[i] = i;
		slider->m_hash2[i] = i;
		slider->m_hash3[i] = i;
	}

	for (i = 0; i < slider->m_V * slider->m_V; ++i)
	{
		int x, y, t;

		x = random() % slider->m_V;
		y = random() % slider->m_V;

		t = slider->m_hash1[x];
		slider->m_hash1[x] = slider->m_hash1[y];
		slider->m_hash1[y] = t;
	}

	for (i = 0; i < slider->m_V * slider->m_V; ++i)
	{
		int x, y, t;

		x = random() % slider->m_V;
		y = random() % slider->m_V;

		t = slider->m_hash2[x];
		slider->m_hash2[x] = slider->m_hash2[y];
		slider->m_hash2[y] = t;
	}

	for (i = 0; i < slider->m_V * slider->m_V; ++i)
	{
		int x, y, t;

		x = random() % slider->m_V;
		y = random() % slider->m_V;

		t = slider->m_hash3[x];
		slider->m_hash3[x] = slider->m_hash3[y];
		slider->m_hash3[y] = t;
	}
}


void destroy_slider(SLIDER *slider)
{
	free(slider->m_cycle3);
	free(slider->m_cycle2);
	free(slider->m_cycle1);
	free(slider->m_hash1);
	free(slider->m_hash2);
	free(slider->m_hash3);
	free(slider->m_goal);
	free(slider->m_vrtxs);
}


void copy_slider(SLIDER *slider, SLIDER *copy)
{
	copy->m_V = slider->m_V;
	copy->m_vrtxs = (int*)malloc(slider->m_V * sizeof(int));
	assert(copy->m_vrtxs != NULL);

	copy->m_goal = (int*)malloc(slider->m_V * sizeof(int));
	assert(copy->m_goal != NULL);

	copy->m_C1 = slider->m_C1;
	copy->m_cycle1 = (int*)malloc(slider->m_C1 * sizeof(int));
	assert(copy->m_cycle1 != NULL);

	copy->m_C2 = slider->m_C2;
	copy->m_cycle2 = (int*)malloc(slider->m_C2 * sizeof(int));
	assert(copy->m_cycle2 != NULL);

	copy->m_C3 = slider->m_C3;
	copy->m_cycle3 = (int*)malloc(slider->m_C3 * sizeof(int));
	assert(copy->m_cycle3 != NULL);

	copy->m_H1 = slider->m_H1;
	copy->m_hash1 = (int*)malloc(slider->m_H1 * sizeof(int));
	assert(copy->m_hash1 != NULL);

	copy->m_H2 = slider->m_H2;
	copy->m_hash2 = (int*)malloc(slider->m_H2 * sizeof(int));
	assert(copy->m_hash2 != NULL);

	copy->m_H3 = slider->m_H3;
	copy->m_hash3 = (int*)malloc(slider->m_H3 * sizeof(int));
	assert(copy->m_hash3 != NULL);

	int i;

	for (i = 0; i < slider->m_V; ++i)
	{
		copy->m_vrtxs[i] = slider->m_vrtxs[i];
		copy->m_goal[i] = slider->m_goal[i];
		copy->m_hash1[i] = slider->m_hash1[i];
		copy->m_hash2[i] = slider->m_hash2[i];
		copy->m_hash3[i] = slider->m_hash3[i];
	}

	for (i = 0; i < slider->m_C1; ++i)
	{
		copy->m_cycle1[i] = slider->m_cycle1[i];
	}
	for (i = 0; i < slider->m_C2; ++i)
	{
		copy->m_cycle2[i] = slider->m_cycle2[i];
	}
	for (i = 0; i < slider->m_C3; ++i)
	{
		copy->m_cycle3[i] = slider->m_cycle3[i];
	}
}


void swap_slider(SLIDER *slider)
{
	int i, t;

	for (i = 0; i < slider->m_V; ++i)
	{
		t = slider->m_goal[i];
		slider->m_goal[i] = slider->m_vrtxs[i];
		slider->m_vrtxs[i] = t;
	}
}


void swap_goal(int s1, int s2, SLIDER *slider)
{
	int t;

	t = slider->m_goal[s1];
	slider->m_goal[s1] = slider->m_goal[s2];
	slider->m_goal[s2] = t;
}


int check_goal(SLIDER *slider)
{
	int i;

	for (i = 0; i < slider->m_V; ++i)
	{
		if (slider->m_vrtxs[i] != slider->m_goal[i])
		{
			return 0;
		}
	}

	return 1;
}


int estimate_depth(SLIDER *slider, DIST_MATRIX *dist_matrix)
{
	int i, j;
	int dist = 0;

	for (i = 0; i < slider->m_V; ++i)
	{
		for (j = 0; j < slider->m_V; ++j)
		{
			if (slider->m_vrtxs[i] == slider->m_goal[j])
			{
				break;
			}
		}

		if (dist_matrix->m_dist[i][j] > dist)
		{
			dist = dist_matrix->m_dist[i][j];
		}
	}

	return dist;
}


void print_slider(SLIDER *slider)
{
	printf("Slider:%d\n", slider->m_V);

	int i;

	printf("Cycle 1:");
	for (i = 0; i < slider->m_C1; ++i)
	{
		printf("%d ", slider->m_cycle1[i]);
	}
	printf("\n");

	printf("State 1:");
	for (i = 0; i < slider->m_C1; ++i)
	{
		printf("%d ", slider->m_vrtxs[slider->m_cycle1[i]]);
	}
	printf("\n");

	printf(" Goal 1:");
	for (i = 0; i < slider->m_C1; ++i)
	{
		printf("%d ", slider->m_goal[slider->m_cycle1[i]]);
	}
	printf("\n");

	printf("Cycle 2:");
	for (i = 0; i < slider->m_C2; ++i)
	{
		printf("%d ", slider->m_cycle2[i]);
	}
	printf("\n");

	printf("State 2:");
	for (i = 0; i < slider->m_C2; ++i)
	{
		printf("%d ", slider->m_vrtxs[slider->m_cycle2[i]]);
	}
	printf("\n");

	printf(" Goal 2:");
	for (i = 0; i < slider->m_C2; ++i)
	{
		printf("%d ", slider->m_goal[slider->m_cycle2[i]]);
	}
	printf("\n");

	printf("Cycle 3:");
	for (i = 0; i < slider->m_C3; ++i)
	{
		printf("%d ", slider->m_cycle3[i]);
	}
	printf("\n");

	printf("State 3:");
	for (i = 0; i < slider->m_C3; ++i)
	{
		printf("%d ", slider->m_vrtxs[slider->m_cycle3[i]]);
	}
	printf("\n");

	printf(" Goal 3:");
	for (i = 0; i < slider->m_C3; ++i)
	{
		printf("%d ", slider->m_goal[slider->m_cycle3[i]]);
	}
	printf("\n");

	printf(" Hash 1:");
	for (i = 0; i < slider->m_H1; ++i)
	{
		printf("%d ", slider->m_vrtxs[slider->m_hash1[i]]);
	}
	printf("\n");
}


void fprint_slider(SLIDER *slider, FILE *f)
{
	fprintf(f, "Slider:%d\n", slider->m_V);

	int i;

	fprintf(f, "Cycle 1:");
	for (i = 0; i < slider->m_C1; ++i)
	{
		fprintf(f, "%d ", slider->m_cycle1[i]);
	}
	fprintf(f, "\n");

	fprintf(f, "State 1:");
	for (i = 0; i < slider->m_C1; ++i)
	{
		fprintf(f, "%d ", slider->m_vrtxs[slider->m_cycle1[i]]);
	}
	fprintf(f, "\n");

	fprintf(f, " Goal 1:");
	for (i = 0; i < slider->m_C1; ++i)
	{
		fprintf(f, "%d ", slider->m_goal[slider->m_cycle1[i]]);
	}
	fprintf(f, "\n");

	fprintf(f, "Cycle 2:");
	for (i = 0; i < slider->m_C2; ++i)
	{
		fprintf(f, "%d ", slider->m_cycle2[i]);
	}
	fprintf(f, "\n");

	fprintf(f, "State 2:");
	for (i = 0; i < slider->m_C2; ++i)
	{
		fprintf(f, "%d ", slider->m_vrtxs[slider->m_cycle2[i]]);
	}
	fprintf(f, "\n");

	fprintf(f, " Goal 2:");
	for (i = 0; i < slider->m_C2; ++i)
	{
		fprintf(f, "%d ", slider->m_goal[slider->m_cycle2[i]]);
	}
	fprintf(f, "\n");

	fprintf(f, "Cycle 3:");
	for (i = 0; i < slider->m_C3; ++i)
	{
		fprintf(f, "%d ", slider->m_cycle3[i]);
	}
	fprintf(f, "\n");

	fprintf(f, "State 3:");
	for (i = 0; i < slider->m_C3; ++i)
	{
		fprintf(f, "%d ", slider->m_vrtxs[slider->m_cycle3[i]]);
	}
	fprintf(f, "\n");

	fprintf(f, " Goal 3:");
	for (i = 0; i < slider->m_C3; ++i)
	{
		fprintf(f, "%d ", slider->m_goal[slider->m_cycle3[i]]);
	}
	fprintf(f, "\n");
}


void cycle_1_rotate_L(SLIDER *slider)
{
	int i, t;

	t = slider->m_vrtxs[slider->m_cycle1[0]];

	for (i = 1; i < slider->m_C1; ++i)
	{
		slider->m_vrtxs[slider->m_cycle1[i - 1]] = slider->m_vrtxs[slider->m_cycle1[i]];
	}
	slider->m_vrtxs[slider->m_cycle1[slider->m_C1 - 1]] = t;

	t = slider->m_vrtxs[slider->m_cycle1[slider->m_C1 - 1]];
	slider->m_vrtxs[slider->m_cycle1[slider->m_C1 - 1]] = slider->m_vrtxs[slider->m_cycle1[0]];
	slider->m_vrtxs[slider->m_cycle1[0]] = t;
}


void cycle_1_rotate_R(SLIDER *slider)
{
	int i, t;

	t = slider->m_vrtxs[slider->m_cycle1[slider->m_C1 - 1]];
	
	for (i = slider->m_C1 - 1; i > 0; --i)
	{
		slider->m_vrtxs[slider->m_cycle1[i]] = slider->m_vrtxs[slider->m_cycle1[i - 1]];
	}
	slider->m_vrtxs[slider->m_cycle1[0]] = t;

	t = slider->m_vrtxs[slider->m_cycle1[1]];
	slider->m_vrtxs[slider->m_cycle1[1]] = slider->m_vrtxs[slider->m_cycle1[0]];
	slider->m_vrtxs[slider->m_cycle1[0]] = t;
}


void cycle_2_rotate_L(SLIDER *slider)
{
	int i, t;

	t = slider->m_vrtxs[slider->m_cycle2[0]];

	for (i = 1; i < slider->m_C2; ++i)
	{
		slider->m_vrtxs[slider->m_cycle2[i - 1]] = slider->m_vrtxs[slider->m_cycle2[i]];
	}
	slider->m_vrtxs[slider->m_cycle2[slider->m_C2 - 1]] = t;

	t = slider->m_vrtxs[slider->m_cycle2[slider->m_C2 - 1]];
	slider->m_vrtxs[slider->m_cycle2[slider->m_C2 - 1]] = slider->m_vrtxs[slider->m_cycle2[0]];
	slider->m_vrtxs[slider->m_cycle2[0]] = t;
}


void cycle_2_rotate_R(SLIDER *slider)
{
	int i, t;

	t = slider->m_vrtxs[slider->m_cycle2[slider->m_C2 - 1]];
	
	for (i = slider->m_C2 - 1; i > 0; --i)
	{
		slider->m_vrtxs[slider->m_cycle2[i]] = slider->m_vrtxs[slider->m_cycle2[i - 1]];
	}
	slider->m_vrtxs[slider->m_cycle2[0]] = t;

	t = slider->m_vrtxs[slider->m_cycle2[1]];
	slider->m_vrtxs[slider->m_cycle2[1]] = slider->m_vrtxs[slider->m_cycle2[0]];
	slider->m_vrtxs[slider->m_cycle2[0]] = t;
}


void cycle_3_rotate_L(SLIDER *slider)
{
	int i, t;

	t = slider->m_vrtxs[slider->m_cycle3[0]];

	for (i = 1; i < slider->m_C3; ++i)
	{
		slider->m_vrtxs[slider->m_cycle3[i - 1]] = slider->m_vrtxs[slider->m_cycle3[i]];
	}
	slider->m_vrtxs[slider->m_cycle3[slider->m_C3 - 1]] = t;


	t = slider->m_vrtxs[slider->m_cycle3[slider->m_C3 - 1]];
	slider->m_vrtxs[slider->m_cycle3[slider->m_C3 - 1]] = slider->m_vrtxs[slider->m_cycle3[0]];
	slider->m_vrtxs[slider->m_cycle3[0]] = t;
}


void cycle_3_rotate_R(SLIDER *slider)
{
	int i, t;

	t = slider->m_vrtxs[slider->m_cycle3[slider->m_C3 - 1]];
	
	for (i = slider->m_C3 - 1; i > 0; --i)
	{
		slider->m_vrtxs[slider->m_cycle3[i]] = slider->m_vrtxs[slider->m_cycle3[i - 1]];
	}
	slider->m_vrtxs[slider->m_cycle3[0]] = t;

	t = slider->m_vrtxs[slider->m_cycle3[1]];
	slider->m_vrtxs[slider->m_cycle3[1]] = slider->m_vrtxs[slider->m_cycle3[0]];
	slider->m_vrtxs[slider->m_cycle3[0]] = t;
}


void init_solution(int max_moves, SOLUTION *solution)
{
	solution->m_maxM = max_moves;
	solution->m_M = 0;

	solution->m_moves = (MOVE*)malloc(solution->m_maxM * sizeof(MOVE));
	assert(solution->m_moves != NULL);
}


void reset_solution(SOLUTION *solution)
{
	solution->m_M = 0;
	solution->m_moves[0] = MOVE_UNDEFINED;
}


void push_move(MOVE move, SOLUTION *solution)
{
	assert(solution->m_M < solution->m_maxM);
	solution->m_moves[solution->m_M++] = move;
	++STEPS;
}


void pop_move(SOLUTION *solution)
{
	--solution->m_M;
}


void print_solution(SOLUTION *solution)
{
	int i;

	printf("Solution:%d\n", solution->m_M);
	
	for (i = 0; i < solution->m_M; ++i)
	{
		switch (solution->m_moves[i])
		{
		case CYCLE_1_LEFT:
		{
			printf("%d:\t1 cw (%d)\n", i, solution->m_moves[i]);
			break;
		}
		case CYCLE_1_RIGHT:
		{
			printf("%d:\t1 ccw (%d)\n", i, solution->m_moves[i]);
			break;
		}
		case CYCLE_2_LEFT:
		{
			printf("%d:\t2 CW (%d)\n", i, solution->m_moves[i]);
			break;
		}
		case CYCLE_2_RIGHT:
		{
			printf("%d:\t2 CCW (%d)\n", i, solution->m_moves[i]);
			break;
		}
		case CYCLE_3_LEFT:
		{
			printf("%d:\t3 cW (%d)\n", i, solution->m_moves[i]);
			break;
		}
		case CYCLE_3_RIGHT:
		{
			printf("%d:\t3 cCW (%d)\n", i, solution->m_moves[i]);
			break;
		}
		default:
		{
			assert(0);
			break;
		}
		}
	}
}


void fprint_solution(SOLUTION *solution, FILE *f)
{
	int i;

	fprintf(f, "Solution:%d\n", solution->m_M);
	
	for (i = 0; i < solution->m_M; ++i)
	{
		switch (solution->m_moves[i])
		{
		case CYCLE_1_LEFT:
		{
			fprintf(f, "%d:\t1 cw\n", i);
			break;
		}
		case CYCLE_1_RIGHT:
		{
			fprintf(f, "%d:\t1 ccw\n", i);
			break;
		}
		case CYCLE_2_LEFT:
		{
			fprintf(f, "%d:\t2 CW\n", i);
			break;
		}
		case CYCLE_2_RIGHT:
		{
			fprintf(f, "%d:\t2 CCW\n", i);
			break;
		}
		case CYCLE_3_LEFT:
		{
			fprintf(f, "%d:\t3 cW\n", i);
			break;
		}
		case CYCLE_3_RIGHT:
		{
			fprintf(f, "%d:\t3 cCW\n", i);
			break;
		}
		default:
		{
			assert(0);
			break;
		}
		}
	}
}


int check_symmetry(SOLUTION *solution)
{
	if (solution->m_M > 1)
	{
		int mid = solution->m_M - 1;
		
		while (mid >= solution->m_M / 2)
		{
			int symmetric = 1;
			
			int i = mid - 1;
			int j = mid;
		
			while (1)
			{
				int s = 0;
				
				if (   (solution->m_moves[i] == CYCLE_1_LEFT  && solution->m_moves[j] == CYCLE_1_RIGHT)
				    || (solution->m_moves[i] == CYCLE_1_RIGHT && solution->m_moves[j] == CYCLE_1_LEFT))
				{
					s = 1;
				}
				else
				{
					if (   (solution->m_moves[i] == CYCLE_2_LEFT  && solution->m_moves[j] == CYCLE_2_RIGHT)
					       || (solution->m_moves[i] == CYCLE_2_RIGHT && solution->m_moves[j] == CYCLE_2_LEFT))
					{
						s = 1;
					}
					else
					{
						if (   (solution->m_moves[i] == CYCLE_3_LEFT  && solution->m_moves[j] == CYCLE_3_RIGHT)
						       || (solution->m_moves[i] == CYCLE_3_RIGHT && solution->m_moves[j] == CYCLE_3_LEFT))
						{
							s = 1;
						}
					}
				}
				
				if (!s)
				{
					symmetric = 0;
					break;
				}
				
				--i;
				++j;
				
				if (i < 0)
				{
					if (j >= solution->m_M)
					{
						break;
					}
					else
					{
						symmetric = 0;
						break;
					}
				}
				else
				{
					if (j >= solution->m_M)
					{
						symmetric = 0;
						break;
					}
				}
			}
			if (symmetric)
			{
				return 0;
			}
			--mid;
		}
		return 1;
	}
	else
	{
		return 1;
	}

//	return 1;
}


int alpha_count = 0;

int solve_slider(int depth, int max_depth, SLIDER *slider, DIST_MATRIX *dist_matrix, SOLUTION *solution)
{
//	print_slider(slider);
/*
	if (depth == max_depth)
	{
		++alpha_count;
		printf("Alpha 1: %d\n", alpha_count);
	}
*/
//	if (depth == max_depth)
	{
		int o_sig = calc_signature(ORIGINAL_1.m_vrtxs, slider->m_vrtxs, slider->m_V);
/*
	printf("     Original signature: %d\n", o_sig);
*/

/*
	printf("         Goal signature: %d\n", g_sig);
*/

		BI_SIGNATURE o_bi_sig;
		calc_bi_signature(slider, &ORIGINAL_1, &o_bi_sig);
/*
	printf("  Original bi-signature: %d,%d\n", o_bi_sig.m_c1_sign, o_bi_sig.m_c2_sign);
*/
/*
	printf("      Goal bi-signature: %d,%d\n", g_bi_sig.m_c1_sign, g_bi_sig.m_c2_sign);
	printf("\n");
*/

		TRI_SIGNATURE o_tri_sig;
		calc_tri_signature(slider, &ORIGINAL_1, &o_tri_sig);

//	printf(" Original tri-signature: %d,%d,%d\n", o_tri_sig.m_c1_sign, o_tri_sig.m_c2_sign, o_tri_sig.m_c3_sign);
//	printf("     Goal tri-signature: %d,%d,%d\n", g_tri_sig.m_c1_sign, g_tri_sig.m_c2_sign, g_tri_sig.m_c3_sign);

		QUAD_SIGNATURE o_quad_sig;
		calc_quad_signature(slider, &ORIGINAL_1, &o_quad_sig);
//		printf("Original quad-signature: %d,%d,%d,%d\n", o_quad_sig.m_c1_sign, o_quad_sig.m_c2_sign, o_quad_sig.m_c3_sign, o_quad_sig.m_c4_sign);

//	printf("    Goal quad-signature: %d,%d,%d,%d\n", g_quad_sig.m_c1_sign, g_quad_sig.m_c2_sign, g_quad_sig.m_c3_sign, g_quad_sig.m_c4_sign);
//	printf("\n");


		if (SIGN_DEPTHS[o_sig] > depth)
		{
			SIGN_DEPTHS[o_sig] = depth;
		}
		
		if (BI_SIGN_DEPTHS[o_bi_sig.m_c1_sign][o_bi_sig.m_c2_sign] > depth)
		{
			BI_SIGN_DEPTHS[o_bi_sig.m_c1_sign][o_bi_sig.m_c2_sign] = depth;
		}

		if (TRI_SIGN_DEPTHS[o_tri_sig.m_c1_sign][o_tri_sig.m_c2_sign][o_tri_sig.m_c3_sign] > depth)
		{
			TRI_SIGN_DEPTHS[o_tri_sig.m_c1_sign][o_tri_sig.m_c2_sign][o_tri_sig.m_c3_sign] = depth;
		}
		
		QUAD_SIGN_DEPTHS_map::iterator o_sign_iter = QUAD_SIGN_DEPTHS.find(o_quad_sig);
		
		if (o_sign_iter == QUAD_SIGN_DEPTHS.end())
		{
			QUAD_SIGN_DEPTHS[o_quad_sig] = depth;
		}
		else
		{
			if (o_sign_iter->second > depth)
			{
				o_sign_iter->second = depth;
			}
		}
	}

	int estimated, estimated_1, estimated_2, estimated_3, estimated_4, estimated_5;

	estimated_1 = estimate_depth(slider, dist_matrix) - 1;

	int g_sig = calc_signature(slider->m_vrtxs, slider->m_goal, slider->m_V);

	estimated_2 = (SIGN_DEPTHS[g_sig] == INT_MAX) ?
		max_depth / 2 + 1: SIGN_DEPTHS[g_sig];

	BI_SIGNATURE g_bi_sig;
	calc_bi_signature(slider, slider, &g_bi_sig);

	estimated_3 = (BI_SIGN_DEPTHS[g_bi_sig.m_c1_sign][g_bi_sig.m_c2_sign] == INT_MAX) ?
		max_depth / 2 + 1: BI_SIGN_DEPTHS[g_bi_sig.m_c1_sign][g_bi_sig.m_c2_sign];

	TRI_SIGNATURE g_tri_sig;
	calc_tri_signature(slider, slider, &g_tri_sig);

	estimated_4 = (TRI_SIGN_DEPTHS[g_tri_sig.m_c1_sign][g_tri_sig.m_c2_sign][g_tri_sig.m_c3_sign] == INT_MAX) ?
		max_depth / 2 + 1: TRI_SIGN_DEPTHS[g_tri_sig.m_c1_sign][g_tri_sig.m_c2_sign][g_tri_sig.m_c3_sign];

	QUAD_SIGNATURE g_quad_sig;
	calc_quad_signature(slider, slider, &g_quad_sig);

	QUAD_SIGN_DEPTHS_map::iterator g_sign_iter = QUAD_SIGN_DEPTHS.find(g_quad_sig);

	if (g_sign_iter == QUAD_SIGN_DEPTHS.end())
	{
		estimated_5 = max_depth / 2 + 1;
	}
	else
	{
		estimated_5 = g_sign_iter->second;
	}

/*
	printf("Estimated 1: %d\n", estimated_1);
	printf("Estimated 2: %d\n", estimated_2);

	if (estimated_2 > estimated_1)
	{
		printf("---------------->\n");
	}
	printf("\n");
*/

	estimated = 0;
//	estimated = (estimated_1 > estimated) ? estimated_1 : estimated;
//	estimated = (estimated_2 > estimated) ? estimated_2 : estimated;
//	estimated = (estimated_3 > estimated) ? estimated_3 : estimated;
//	estimated = (estimated_4 > estimated) ? estimated_4 : estimated;

//	printf("Size: %d\n", QUAD_SIGN_DEPTHS.size());
/*
	if (estimated < estimated_5)
	{
		printf("<----------------************---------------: %d x %d\n", estimated, estimated_5);
	}
*/
	estimated = (estimated_5 > estimated) ? estimated_5 : estimated;

	if (depth <= max_depth / 2)
	{
		if (estimated > 0)
		{
			--estimated;
		}
	}

	if (depth + estimated < max_depth)
//	if (depth + estimated <= 13 && depth <= max_depth)
	{
/*
		int estimated_1L, estimated_1R, estimated_2L, estimated_2R, estimated_3L, estimated_3R;

		cycle_1_rotate_L(slider);
		QUAD_SIGNATURE r_quad_sig;
		calc_quad_signature(slider, slider, &r_quad_sig);

		QUAD_SIGN_DEPTHS_map::iterator r_sign_iter = QUAD_SIGN_DEPTHS.find(r_quad_sig);

		if (r_sign_iter == QUAD_SIGN_DEPTHS.end())
		{
			estimated_1L = max_depth / 2 + 1;
		}
		else
		{
			estimated_1L = r_sign_iter->second;
		}
		cycle_1_rotate_R(slider);

		cycle_1_rotate_R(slider);
		calc_quad_signature(slider, slider, &r_quad_sig);

		r_sign_iter = QUAD_SIGN_DEPTHS.find(r_quad_sig);

		if (r_sign_iter == QUAD_SIGN_DEPTHS.end())
		{
			estimated_1R = max_depth / 2 + 1;
		}
		else
		{
			estimated_1R = r_sign_iter->second;
		}
		cycle_1_rotate_L(slider);

		cycle_2_rotate_L(slider);
		calc_quad_signature(slider, slider, &r_quad_sig);

		r_sign_iter = QUAD_SIGN_DEPTHS.find(r_quad_sig);

		if (r_sign_iter == QUAD_SIGN_DEPTHS.end())
		{
			estimated_2L = max_depth / 2 + 1;
		}
		else
		{
			estimated_2L = r_sign_iter->second;
		}
		cycle_2_rotate_R(slider);

		cycle_2_rotate_R(slider);
		calc_quad_signature(slider, slider, &r_quad_sig);

		r_sign_iter = QUAD_SIGN_DEPTHS.find(r_quad_sig);

		if (r_sign_iter == QUAD_SIGN_DEPTHS.end())
		{
			estimated_2R = max_depth / 2 + 1;
		}
		else
		{
			estimated_2R = r_sign_iter->second;
		}
		cycle_2_rotate_L(slider);

		cycle_3_rotate_L(slider);
		calc_quad_signature(slider, slider, &r_quad_sig);

		r_sign_iter = QUAD_SIGN_DEPTHS.find(r_quad_sig);

		if (r_sign_iter == QUAD_SIGN_DEPTHS.end())
		{
			estimated_3L = max_depth / 2 + 1;
		}
		else
		{
			estimated_3L = r_sign_iter->second;
		}
		cycle_3_rotate_R(slider);

		cycle_3_rotate_R(slider);
		calc_quad_signature(slider, slider, &r_quad_sig);

		r_sign_iter = QUAD_SIGN_DEPTHS.find(r_quad_sig);

		if (r_sign_iter == QUAD_SIGN_DEPTHS.end())
		{
			estimated_3R = max_depth / 2 + 1;
		}
		else
		{
			estimated_3R = r_sign_iter->second;
		}
		cycle_3_rotate_L(slider);

		printf("%d,%d,%d,%d,%d,%d\n", estimated_1L, estimated_1R, estimated_2L, estimated_2R, estimated_3L, estimated_3R);
		
		int moves = 0, tresh = 0;
*/
//		while (moves < 6)
		{
//			if (tresh == estimated_1L)
			{
				push_move(CYCLE_1_LEFT, solution);
				if (check_symmetry(solution))
				{
					cycle_1_rotate_L(slider);
					
					if (check_goal(slider))
					{
						return 1;
					}
					else
					{
						if (solve_slider(depth + 1, max_depth, slider, dist_matrix, solution))
						{
							return 1;
						}
						else
						{
							cycle_1_rotate_R(slider);
						}
					}
				}
				pop_move(solution);
//				++moves;
			}
//			if (tresh == estimated_1R)
			{
				push_move(CYCLE_1_RIGHT, solution);
				if (check_symmetry(solution))
				{
					cycle_1_rotate_R(slider);
					
					if (check_goal(slider))
					{
						return 1;
					}
					else
					{
						if (solve_slider(depth + 1, max_depth, slider, dist_matrix, solution))
						{
							return 1;
						}
						else
						{
							cycle_1_rotate_L(slider);
						}
					}
				}
				pop_move(solution);
//				++moves;
			}
//			if (tresh == estimated_2L)
			{				
				push_move(CYCLE_2_LEFT, solution);
				if (check_symmetry(solution))
				{
					cycle_2_rotate_L(slider);
					
					if (check_goal(slider))
					{
						return 1;
					}
					else
					{
						if (solve_slider(depth + 1, max_depth, slider, dist_matrix, solution))
						{
							return 1;
						}
						else
						{
							cycle_2_rotate_R(slider);
						}
					}
				}
				pop_move(solution);
//				++moves;
			}
//			if (tresh == estimated_2R)
			{
				push_move(CYCLE_2_RIGHT, solution);
				if (check_symmetry(solution))
				{
					cycle_2_rotate_R(slider);
					
					if (check_goal(slider))
					{
						return 1;
					}
					else
					{
						if (solve_slider(depth + 1, max_depth, slider, dist_matrix, solution))
						{
							return 1;
						}
						else
						{
							cycle_2_rotate_L(slider);
						}
					}
				}
				pop_move(solution);
//				++moves;
			}
//			if (tresh == estimated_3L)
			{
				push_move(CYCLE_3_LEFT, solution);
				if (check_symmetry(solution))
				{
					cycle_3_rotate_L(slider);
					
					if (check_goal(slider))
					{
						return 1;
					}
					else
					{
						if (solve_slider(depth + 1, max_depth, slider, dist_matrix, solution))
						{
							return 1;
						}
						else
						{
							cycle_3_rotate_R(slider);
						}
					}
				}
				pop_move(solution);
//				++moves;
			}
//			if (tresh == estimated_3R)
			{
				push_move(CYCLE_3_RIGHT, solution);
				if (check_symmetry(solution))
				{
					cycle_3_rotate_R(slider);
					
					if (check_goal(slider))
					{
						return 1;
					}
					else
					{
						if (solve_slider(depth + 1, max_depth, slider, dist_matrix, solution))
						{
							return 1;
						}
						else
						{
							cycle_3_rotate_L(slider);
						}
					}
				}
				pop_move(solution);
//				++moves;
			}
//			++tresh;
		}

		return 0;
	}
	else
	{
		return 0;
	}
}


int solve_slider_R(int depth, int max_depth, SLIDER *slider, DIST_MATRIX *dist_matrix, SOLUTION *solution)
{
//	print_slider(slider);
/*
	if (depth == max_depth)
	{
		++alpha_count;
		printf("Alpha 1: %d\n", alpha_count);
	}
*/
//	if (depth == max_depth)
	{
//		int o_sig = calc_signature(ORIGINAL.m_vrtxs, slider->m_vrtxs, slider->m_V);
/*
	printf("     Original signature: %d\n", o_sig);
*/

/*
	printf("         Goal signature: %d\n", g_sig);
*/
/*

		BI_SIGNATURE o_bi_sig;
		calc_bi_signature(slider, &ORIGINAL, &o_bi_sig);
*/
/*
	printf("  Original bi-signature: %d,%d\n", o_bi_sig.m_c1_sign, o_bi_sig.m_c2_sign);
*/
/*
	printf("      Goal bi-signature: %d,%d\n", g_bi_sig.m_c1_sign, g_bi_sig.m_c2_sign);
	printf("\n");
*/
/*
		TRI_SIGNATURE o_tri_sig;
		calc_tri_signature(slider, &ORIGINAL, &o_tri_sig);
*/
//	printf(" Original tri-signature: %d,%d,%d\n", o_tri_sig.m_c1_sign, o_tri_sig.m_c2_sign, o_tri_sig.m_c3_sign);
//	printf("     Goal tri-signature: %d,%d,%d\n", g_tri_sig.m_c1_sign, g_tri_sig.m_c2_sign, g_tri_sig.m_c3_sign);

		QUAD_SIGNATURE o_quad_sig, r_quad_sig;
		calc_quad_signature(slider, &ORIGINAL_2, &o_quad_sig);

		calc_quad_signature(slider, &ORIGINAL_1, &r_quad_sig);
		printf("--> R <-- Original quad-signature: %d,%d,%d,%d\n", r_quad_sig.m_c1_sign, r_quad_sig.m_c2_sign, r_quad_sig.m_c3_sign, r_quad_sig.m_c4_sign);

//	printf("    Goal quad-signature: %d,%d,%d,%d\n", g_quad_sig.m_c1_sign, g_quad_sig.m_c2_sign, g_quad_sig.m_c3_sign, g_quad_sig.m_c4_sign);
//	printf("\n");

/*
		if (SIGN_DEPTHS[o_sig] > depth)
		{
			SIGN_DEPTHS[o_sig] = depth;
		}
		
		if (BI_SIGN_DEPTHS[o_bi_sig.m_c1_sign][o_bi_sig.m_c2_sign] > depth)
		{
			BI_SIGN_DEPTHS[o_bi_sig.m_c1_sign][o_bi_sig.m_c2_sign] = depth;
		}

		if (TRI_SIGN_DEPTHS[o_tri_sig.m_c1_sign][o_tri_sig.m_c2_sign][o_tri_sig.m_c3_sign] > depth)
		{
			TRI_SIGN_DEPTHS[o_tri_sig.m_c1_sign][o_tri_sig.m_c2_sign][o_tri_sig.m_c3_sign] = depth;
		}
*/	
		QUAD_SIGN_DEPTHS_map::iterator o_sign_iter = QUAD_SIGN_R_DEPTHS.find(o_quad_sig);
		
		if (o_sign_iter == QUAD_SIGN_R_DEPTHS.end())
		{
			QUAD_SIGN_R_DEPTHS[o_quad_sig] = depth;
		}
		else
		{
			if (o_sign_iter->second > depth)
			{
				o_sign_iter->second = depth;
			}
		}

		QUAD_SIGN_DEPTHS_map::iterator r_sign_iter = QUAD_SIGN_RR_DEPTHS.find(r_quad_sig);
		
		if (r_sign_iter == QUAD_SIGN_RR_DEPTHS.end())
		{
			QUAD_SIGN_RR_DEPTHS[r_quad_sig] = depth;
		}
		else
		{
			if (r_sign_iter->second > depth)
			{
				r_sign_iter->second = depth;
			}
		}
	}

	int estimated, estimated_1, estimated_5;

//	estimated_1 = estimate_depth(slider, dist_matrix) - 1;

	QUAD_SIGNATURE g_quad_sig;
	calc_quad_signature(slider, slider, &g_quad_sig);

	QUAD_SIGN_DEPTHS_map::iterator g_sign_iter = QUAD_SIGN_R_DEPTHS.find(g_quad_sig);

	if (g_sign_iter == QUAD_SIGN_R_DEPTHS.end())
	{
		estimated_5 = max_depth / 2 + 1;
	}
	else
	{
		estimated_5 = g_sign_iter->second;
	}

/*
	printf("Estimated 1: %d\n", estimated_1);
	printf("Estimated 2: %d\n", estimated_2);

	if (estimated_2 > estimated_1)
	{
		printf("---------------->\n");
	}
	printf("\n");
*/

	estimated = 0;
//	estimated = (estimated_1 > estimated) ? estimated_1 : estimated;
//	estimated = (estimated_2 > estimated) ? estimated_2 : estimated;
//	estimated = (estimated_3 > estimated) ? estimated_3 : estimated;
//	estimated = (estimated_4 > estimated) ? estimated_4 : estimated;

//	printf("Size: %d\n", QUAD_SIGN_DEPTHS.size());
/*
	if (estimated < estimated_5)
	{
		printf("<----------------************---------------: %d x %d\n", estimated, estimated_5);
	}
*/
	estimated = (estimated_5 > estimated) ? estimated_5 : estimated;

	if (depth <= max_depth / 2)
	{
		if (estimated > 0)
		{
			--estimated;
		}
	}

	if (depth + estimated < max_depth)
//	if (depth + estimated <= 13 && depth <= max_depth)
	{
/*
		int estimated_1L, estimated_1R, estimated_2L, estimated_2R, estimated_3L, estimated_3R;

		cycle_1_rotate_L(slider);
		QUAD_SIGNATURE r_quad_sig;
		calc_quad_signature(slider, slider, &r_quad_sig);

		QUAD_SIGN_DEPTHS_map::iterator r_sign_iter = QUAD_SIGN_DEPTHS.find(r_quad_sig);

		if (r_sign_iter == QUAD_SIGN_DEPTHS.end())
		{
			estimated_1L = max_depth / 2 + 1;
		}
		else
		{
			estimated_1L = r_sign_iter->second;
		}
		cycle_1_rotate_R(slider);

		cycle_1_rotate_R(slider);
		calc_quad_signature(slider, slider, &r_quad_sig);

		r_sign_iter = QUAD_SIGN_DEPTHS.find(r_quad_sig);

		if (r_sign_iter == QUAD_SIGN_DEPTHS.end())
		{
			estimated_1R = max_depth / 2 + 1;
		}
		else
		{
			estimated_1R = r_sign_iter->second;
		}
		cycle_1_rotate_L(slider);

		cycle_2_rotate_L(slider);
		calc_quad_signature(slider, slider, &r_quad_sig);

		r_sign_iter = QUAD_SIGN_DEPTHS.find(r_quad_sig);

		if (r_sign_iter == QUAD_SIGN_DEPTHS.end())
		{
			estimated_2L = max_depth / 2 + 1;
		}
		else
		{
			estimated_2L = r_sign_iter->second;
		}
		cycle_2_rotate_R(slider);

		cycle_2_rotate_R(slider);
		calc_quad_signature(slider, slider, &r_quad_sig);

		r_sign_iter = QUAD_SIGN_DEPTHS.find(r_quad_sig);

		if (r_sign_iter == QUAD_SIGN_DEPTHS.end())
		{
			estimated_2R = max_depth / 2 + 1;
		}
		else
		{
			estimated_2R = r_sign_iter->second;
		}
		cycle_2_rotate_L(slider);

		cycle_3_rotate_L(slider);
		calc_quad_signature(slider, slider, &r_quad_sig);

		r_sign_iter = QUAD_SIGN_DEPTHS.find(r_quad_sig);

		if (r_sign_iter == QUAD_SIGN_DEPTHS.end())
		{
			estimated_3L = max_depth / 2 + 1;
		}
		else
		{
			estimated_3L = r_sign_iter->second;
		}
		cycle_3_rotate_R(slider);

		cycle_3_rotate_R(slider);
		calc_quad_signature(slider, slider, &r_quad_sig);

		r_sign_iter = QUAD_SIGN_DEPTHS.find(r_quad_sig);

		if (r_sign_iter == QUAD_SIGN_DEPTHS.end())
		{
			estimated_3R = max_depth / 2 + 1;
		}
		else
		{
			estimated_3R = r_sign_iter->second;
		}
		cycle_3_rotate_L(slider);

		printf("%d,%d,%d,%d,%d,%d\n", estimated_1L, estimated_1R, estimated_2L, estimated_2R, estimated_3L, estimated_3R);
		
		int moves = 0, tresh = 0;
*/
//		while (moves < 6)
		{
//			if (tresh == estimated_1L)
			{
				push_move(CYCLE_1_LEFT, solution);
				if (check_symmetry(solution))
				{
					cycle_1_rotate_L(slider);
					
					if (check_goal(slider))
					{
						return 1;
					}
					else
					{
						if (solve_slider_R(depth + 1, max_depth, slider, dist_matrix, solution))
						{
							return 1;
						}
						else
						{
							cycle_1_rotate_R(slider);
						}
					}
				}
				pop_move(solution);
//				++moves;
			}
//			if (tresh == estimated_1R)
			{
				push_move(CYCLE_1_RIGHT, solution);
				if (check_symmetry(solution))
				{
					cycle_1_rotate_R(slider);
					
					if (check_goal(slider))
					{
						return 1;
					}
					else
					{
						if (solve_slider_R(depth + 1, max_depth, slider, dist_matrix, solution))
						{
							return 1;
						}
						else
						{
							cycle_1_rotate_L(slider);
						}
					}
				}
				pop_move(solution);
//				++moves;
			}
//			if (tresh == estimated_2L)
			{				
				push_move(CYCLE_2_LEFT, solution);
				if (check_symmetry(solution))
				{
					cycle_2_rotate_L(slider);
					
					if (check_goal(slider))
					{
						return 1;
					}
					else
					{
						if (solve_slider_R(depth + 1, max_depth, slider, dist_matrix, solution))
						{
							return 1;
						}
						else
						{
							cycle_2_rotate_R(slider);
						}
					}
				}
				pop_move(solution);
//				++moves;
			}
//			if (tresh == estimated_2R)
			{
				push_move(CYCLE_2_RIGHT, solution);
				if (check_symmetry(solution))
				{
					cycle_2_rotate_R(slider);
					
					if (check_goal(slider))
					{
						return 1;
					}
					else
					{
						if (solve_slider_R(depth + 1, max_depth, slider, dist_matrix, solution))
						{
							return 1;
						}
						else
						{
							cycle_2_rotate_L(slider);
						}
					}
				}
				pop_move(solution);
//				++moves;
			}
//			if (tresh == estimated_3L)
			{
				push_move(CYCLE_3_LEFT, solution);
				if (check_symmetry(solution))
				{
					cycle_3_rotate_L(slider);
					
					if (check_goal(slider))
					{
						return 1;
					}
					else
					{
						if (solve_slider_R(depth + 1, max_depth, slider, dist_matrix, solution))
						{
							return 1;
						}
						else
						{
							cycle_3_rotate_R(slider);
						}
					}
				}
				pop_move(solution);
//				++moves;
			}
//			if (tresh == estimated_3R)
			{
				push_move(CYCLE_3_RIGHT, solution);
				if (check_symmetry(solution))
				{
					cycle_3_rotate_R(slider);
					
					if (check_goal(slider))
					{
						return 1;
					}
					else
					{
						if (solve_slider_R(depth + 1, max_depth, slider, dist_matrix, solution))
						{
							return 1;
						}
						else
						{
							cycle_3_rotate_L(slider);
						}
					}
				}
				pop_move(solution);
//				++moves;
			}
//			++tresh;
		}

		return 0;
	}
	else
	{
		return 0;
	}
}


void reset_estimations(void)
{
	int i, j, k;
	
	for (i = 0; i < SIGNS_COUNT; ++i)
	{
		SIGN_DEPTHS[i] = INT_MAX;
	}
	
	for (i = 0; i < BI_SIGNS_COUNT; ++i)
	{
		for (j = 0; j < BI_SIGNS_COUNT; ++j)
		{
			BI_SIGN_DEPTHS[i][j] = INT_MAX;
		}
	}
	
	for (i = 0; i < TRI_SIGNS_COUNT; ++i)
	{
		for (j = 0; j < TRI_SIGNS_COUNT; ++j)
		{
			for (k = 0; k < TRI_SIGNS_COUNT; ++k)
			{
				TRI_SIGN_DEPTHS[i][j][k] = INT_MAX;
			}
		}
	}
	
	QUAD_SIGN_DEPTHS.clear();
						
}


int solve_slider_deepening(int id, double time_out, SLIDER *slider1, SLIDER *slider2, SOLUTION *solution1, SOLUTION *solution2, FILE *f)
{
	int i;
	double elapsed;
	clock_t begin, end;

	begin = clock();
	copy_slider(slider1, &ORIGINAL_1);
	copy_slider(slider2, &ORIGINAL_2);

	for (i = 0; i < slider1->m_V; ++i)
	{
		ORIGINAL_1.m_goal[i] = slider1->m_vrtxs[i];
	}
	for (i = 0; i < slider2->m_V; ++i)
	{
		ORIGINAL_2.m_goal[i] = slider2->m_vrtxs[i];
	}

	int max_depth = MAX_DEPTHS[id];
	
//	fprint_slider(slider, f);
	
	DIST_MATRIX dist_matrix1;
	init_dist_matrix(slider1, &dist_matrix1);

	DIST_MATRIX dist_matrix2;
	init_dist_matrix(slider2, &dist_matrix2);

	if (MAX_DEPTHS[id] != INT_MAX)
	{
		do
		{
//			printf("Depth %d ...\n", max_depth);
			reset_solution(solution1);
			int solved1 = solve_slider(0, max_depth, slider1, &dist_matrix1, solution1);
/*
			reset_solution(solution2);
			int solved2 = solve_slider_R(0, max_depth, slider2, &dist_matrix2, solution2);

			QUAD_SIGN_DEPTHS_map::iterator sign_depth;

			for (sign_depth = QUAD_SIGN_DEPTHS.begin(); sign_depth != QUAD_SIGN_DEPTHS.end(); ++sign_depth)
			{
				QUAD_SIGN_DEPTHS_map::iterator r_sign_iter = QUAD_SIGN_RR_DEPTHS.find(sign_depth->first);

				if (r_sign_iter != QUAD_SIGN_RR_DEPTHS.end())
				{
					printf("--------------> Intersection:%d <---------------------\n", max_depth);
				}
			}
*/		
			end = clock();
			elapsed = (end - begin) / (double)CLOCKS_PER_SEC;

			if (solved1)
			{
				MAX_DEPTHS[id] = INT_MAX;
				return 1;
			}
/*
			print_signatures(slider->m_V);
			printf("\n");
			print_bi_signatures(slider->m_V);
*/
			++max_depth;

		} while (elapsed < time_out);

//		fprintf(f, "Time out reached ... elpased:%.3f   steps:%ld   timeout:%.3f   depth reached:%d\n", elapsed, STEPS, time_out, max_depth - 1);
		MAX_DEPTHS[id] = max_depth;
	}
	else
	{
//		fprintf(f, "Already solved\n");
	}

	return 0;
}


void generate_infinity(SLIDER *slider1, SLIDER *slider2, SOLUTION *solution1, SOLUTION *solution2)
{
	int id;
	char filename[1024];

	double time_out = 100.0;

	int c1_b, c2_a, c2_b;

	FILE *fout;

	while (1)
	{
		printf("Computing for timeout:%f...\n", time_out);

		id = 1;

		for (c1_b = 3; c1_b < 16; ++c1_b)
		{
			for (c2_a = 1; c2_a <= c1_b; ++c2_a)
			{
				for (c2_b = c1_b + 1; c1_b - c2_a + c2_b - c1_b + 1 <= c1_b; ++c2_b)
				{
					init_slider(0, c1_b, c2_a, c2_b, slider1);
					int oddness = slider1->m_C1 % 2 || slider1->m_C2 % 2 || slider1->m_C3 % 2;

					clock_t begin, end;

					if (oddness)
					{
						swap_goal(1, 2, slider1);
/*
						copy_slider(slider1, slider2);
						swap_slider(slider2);
*/
						STEPS = 0;

						begin = clock();
						reset_estimations();
						int result = solve_slider_deepening(id, time_out, slider1, slider2, solution1, solution2, fout);
						end = clock();
						double elapsed = (end - begin) / (double)CLOCKS_PER_SEC;

						if (result)
						{
							sprintf(filename, "%d_%d_%d_t_%f___X4.txt", c1_b, c2_a, c2_b, time_out);
							fout = fopen(filename, "w");
							fprintf(fout, "Id:%d\n", id);

							fprintf(fout, "Init:%d %d %d\n", c1_b, c2_a, c2_b);
							fprintf(fout, "Cycs:%d %d %d\n", slider1->m_C1, slider1->m_C2, slider1->m_C3);

							fprint_slider(slider1, fout);
							fprint_solution(solution1, fout);
							fprintf(fout, "Solving time:%.3f   steps:%ld\n", elapsed, STEPS);

							fprintf(fout, "Oddness x solution: %d x %d\n", oddness, result);
							fclose(fout);
						}

					}

					destroy_slider(slider1);
//					destroy_slider(slider2);
					++id;
				}
			}
			++c1_b;
		}

		time_out *= 2;
	}
}


SLIDER slider1, slider2;
SLIDER copy;
SOLUTION solution1;
SOLUTION solution2;


int perm1[] = {5, 2, 3, 4, 1};
int perm2[] = {1, 4, 3, 2, 5};


int main(int argc, char **argv)
{
	int i, j, k;
	clock_t begin, end;
/*
	int sig = calc_signature(perm1, perm2, 5);
	printf("Signature:%d\n", sig);
	exit(0);
*/
	for (i = 0; i < MAX_DEPTHS_COUNT; ++i)
	{
		MAX_DEPTHS[i] = 0;
	}

	reset_estimations();

	FILE *fout;
	fout = fopen("7_6_9___b1.txt", "w");

	begin = clock();

	init_solution(1024, &solution1);
	init_solution(1024, &solution2);
/*
	init_slider(0, 7, 6, 9, &slider1);
	swap_goal(1, 2, &slider1);

	copy_slider(&slider1, &slider2);
	swap_slider(&slider2);

	solve_slider_deepening(48, 1000.0, &slider1, &slider2,  &solution1, &solution2, fout);
	double ratio = analyze_tri_signatures(slider1.m_V);
	printf("Tri signatures ratio: %f\n", ratio);
	fclose(fout);
*/
	generate_infinity(&slider1, &slider2, &solution1, &solution2);

	end = clock();
	printf("Time:%.3f\n", (end - begin) / (double)CLOCKS_PER_SEC);

	return 0;
}
